/* -*- mode: c++ -*- */

#include <stdio.h>
#include <assert.h>
#include <limits.h>
#include "simplegraph.h"
#include "worklist.cuh"
#include "common.h"
#include "Timer.h"
#include <algorithm>
#include <utility>

// SSSP Near--Far, based on Davidson et al., IPDPS 2014
#define DEBUG 0

const int INF = INT_MAX;
const unsigned int WLSIZE = 1000000;
const int DEFAULT_DELTA = 10000;

__global__
void sssp_init(GPUSimpleCSRGraphUII g, unsigned int src) {
  int tid = threadIdx.x + blockDim.x * blockIdx.x;
  int nthreads = gridDim.x * blockDim.x;

  for(int i = tid; i < g.num_nodes; i += nthreads) {
    g.node_wt[i] = (i == src) ? 0 : INF;
  }
}

template <typename T>
__global__ 
void sssp(GPUSimpleCSRGraphUII g, const int delta, Worklist<T> in_wl, Worklist<T> out_wl, Worklist<T> re_wl) {
  int tid = threadIdx.x + blockDim.x * blockIdx.x;
  int nthreads = gridDim.x * blockDim.x;

  for(int i = tid; i < *in_wl.ndx; i += nthreads) {

    bool pop;
    T node;

    pop = in_wl.pop_index(i, node);

    assert(pop);

    for(int e = g.row_start[node]; e < g.row_start[node + 1]; e++) {

      unsigned int dest = g.edge_dst[e];
      unsigned int distance = g.node_wt[node] + g.edge_wt[e];

      int prev_distance = g.node_wt[dest];
      
      if(prev_distance > distance) {
	g.node_wt[dest] = distance;

	if(distance <= delta) {
	  re_wl.push(dest);
	}
	else {
	  out_wl.push(dest);
	}
      }
    }
  }
}

void write_output(SimpleCSRGraphUII &g, const char *out) {
  FILE *fp; 
  
  fp = fopen(out, "w");
  if(!fp) {
    fprintf(stderr, "ERROR: Unable to open output file '%s'\n", out);
    exit(1);
  }

  for(int i = 0; i < g.num_nodes; i++) {
    int r;
    if(g.node_wt[i] == INF) {
      r = fprintf(fp, "%d INF\n", i);
    } else {
      r = fprintf(fp, "%d %d\n", i, g.node_wt[i]);
    }

    if(r < 0) {
      fprintf(stderr, "ERROR: Unable to write output\n");
      exit(1);
    };    
  }
}

int main(int argc, char *argv[]) 
{
  if(argc != 3 && argc != 4) {
    fprintf(stderr, "Usage: %s inputgraph outputgraph [delta]\n", argv[0]);
    exit(1);
  }

  SimpleCSRGraphUII input;

  if(!input.load_file(argv[1])) {
    fprintf(stderr, "ERROR: failed to load graph\n");
    exit(1);
  } 

  printf("Loaded '%s', %u nodes, %u edges\n", argv[1], input.num_nodes, input.num_edges);

  int curdelta = 0, DELTA = DEFAULT_DELTA;

  /* optional argument */
  if(argc == 4) {
    DELTA = atoi(argv[3]);
  }

  printf("DELTA=%d\n", DELTA);

  ggc::Timer t("sssp");

  t.start();

  GPUSimpleCSRGraphUII ginput(input);
  
  Worklist<unsigned int> in_wl(std::max(WLSIZE, (unsigned) input.num_edges)), out_wl(std::max(WLSIZE, (unsigned) input.num_edges));
  Worklist<unsigned int> re_wl(std::max(WLSIZE, (unsigned) input.num_edges));
  
  unsigned int initial[] = {0};
  in_wl.initialize(initial, 1);

  int blocks = (input.num_nodes + 255 / 256); 

  Worklist<unsigned int> *wl[3] = {&in_wl, &out_wl, &re_wl};
  int in_wl_ndx = 0, out_wl_ndx = 1, re_wl_ndx = 2;


  sssp_init<<<blocks, 256>>>(ginput, initial[0]);

  while(wl[in_wl_ndx]->nitems()) {   
    
    while(wl[in_wl_ndx]->nitems()) {
      sssp<<<blocks, 256>>>(ginput, curdelta, *wl[in_wl_ndx], *wl[out_wl_ndx], *wl[re_wl_ndx]);

      wl[in_wl_ndx]->reset();
      std::swap(in_wl_ndx, re_wl_ndx);      
    }

    std::swap(in_wl_ndx, out_wl_ndx);
    curdelta += DELTA;
  }

  in_wl.free();
  out_wl.free();   
  re_wl.free();

  check_cuda(cudaMemcpy(input.node_wt, ginput.node_wt, 
			sizeof(ginput.node_wt[0]) * input.num_nodes,
			cudaMemcpyDeviceToHost));
  t.stop();

  printf("Total time: %u ms\n", t.duration_ms());

  write_output(input, argv[2]);

  printf("Wrote output '%s'\n", argv[2]);
  return 0;
}
